import torch
import torch.nn as nn
import torch.nn.functional as F
import torchaudio.transforms as T


# TDNN code is copied from: https://github.com/cvqluu/TDNN/tree/master
class TdnnLayer(nn.Module):
    def __init__(
                    self, 
                    input_dim=13, 
                    output_dim=128,
                    context_size=5,
                    dilation=1,
                    stride=1,
                    batch_norm=False,
                    dropout_p=0.0
                ):
        super(TdnnLayer, self).__init__()
        self.context_size = context_size
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.dilation = dilation
        self.dropout_p = dropout_p
        self.batch_norm = batch_norm
        self.stride= stride

        self.kernel = nn.Linear(input_dim*context_size, output_dim)
        self.nonlinearity = nn.ReLU()
        if self.batch_norm:
            self.bn = nn.BatchNorm1d(output_dim)
        if self.dropout_p:
            self.drop = nn.Dropout(p=self.dropout_p)
    def forward(self, x):
        '''
        input: size (batch, seq_len, input_features)
        outpu: size (batch, new_seq_len, output_features)
        '''

        _, _, d = x.shape
        assert (d == self.input_dim), 'Input dimension was wrong. Expected ({}), got ({})'.format(self.input_dim, d)
        x = x.unsqueeze(1)

        # Unfold input into smaller temporal contexts
        x = F.unfold(
                        x, 
                        (self.context_size, self.input_dim), 
                        stride=(1,self.input_dim), 
                        dilation=(self.dilation,1)
                    )

        # N, output_dim*context_size, new_t = x.shape
        x = x.transpose(1,2)
        x = self.kernel(x)
        x = self.nonlinearity(x)
        
        if self.dropout_p:
            x = self.drop(x)

        if self.batch_norm:
            x = x.transpose(1,2)
            x = self.bn(x)
            x = x.transpose(1,2)

        return x


class XVectorModel(nn.Module):
    def __init__(self, sample_rate=16000,mfcc_dimensions = 13,dropout_p=0.1,batch_norm=True):
        super(XVectorModel, self).__init__()
        self.output_dim = 256
        self.dims = 128
        self.sample_rate = sample_rate
        # The mfccs with this parameters shorthen the sequence  ~160 time
        self.tdnn1 = TdnnLayer(input_dim=mfcc_dimensions, output_dim=self.dims, context_size=5, dilation=1,dropout_p=dropout_p,batch_norm=batch_norm)
        self.tdnn2 = TdnnLayer(input_dim=self.dims, output_dim=self.dims, context_size=3, dilation=1,dropout_p=dropout_p,batch_norm=batch_norm)
        self.tdnn3 = TdnnLayer(input_dim=self.dims, output_dim=self.dims, context_size=2, dilation=2,dropout_p=dropout_p,batch_norm=batch_norm)
        self.tdnn4 = TdnnLayer(input_dim=self.dims, output_dim=self.dims, context_size=1, dilation=1,dropout_p=dropout_p,batch_norm=batch_norm)
        # with the fifth layer the results were about the same        
        # self.tdnn5 = TdnnLayer(input_dim=self.dims, output_dim=self.dims, context_size=1, dilation=1,dropout_p=dropout_p,batch_norm=batch_norm)



        self.segment_layer1 = nn.Sequential(
            nn.Linear(self.dims*2, self.output_dim),
            nn.ReLU(),
            nn.Dropout(p=dropout_p)
        )
        self.segment_layer2 = nn.Linear(self.output_dim, self.output_dim)


    def forward(self, x):
        data,original_lengths = x
        mfcc_features = data
        # trasnpose [B,MFCC_DIM,SEQ_LEN] into [B,SEQ_LEN,MFCC_DIM] 
        tdnn1_out = self.tdnn1(mfcc_features.transpose(1,2))
        tdnn2_out = self.tdnn2(tdnn1_out)
        tdnn3_out = self.tdnn3(tdnn2_out)
        tdnn4_out = self.tdnn4(tdnn3_out)

        # tdnn5_out = self.tdnn5(tdnn4_out)

        # the output of tddn, is slightly shorter (14), as expected
        # the context os last layer is 15, so last 7 features is contaminated
        # So the index of last good segmenst is same as before -(7+7+7)
        cut_tail = 21
        pooled = self.statistical_pooling(tdnn4_out,original_lengths,cut_tail=cut_tail)
        segment_out = self.segment_layer1(pooled)
        x_vector = self.segment_layer2(segment_out)
        return x_vector
    
    # Performs statistical pooling, but is pooling only from the valid part of input
    def statistical_pooling(self,tdnn_output,original_lengths,cut_tail = 21):
        B = tdnn_output.size(0)
        means = []
        stds = []
        for b in range(B):
            valid_segments_series_length = (original_lengths[b]) - cut_tail
            valid_segment = tdnn_output[b,:valid_segments_series_length,:]
            mean = valid_segment.mean(0)
            std = valid_segment.std(0)
            means.append(mean)
            stds.append(std)
        means = torch.stack(means)
        stds = torch.stack(stds)
        pooled = torch.cat((means,stds),1)
        return pooled