import os
from glob import glob
import torch
from torch.utils.data import Dataset
from torchvision import transforms
from torchvision.io import read_image
import torchaudio
import torchaudio.transforms as T
import random
import torchaudio.functional as F
import torchvision.transforms.functional as TF


class AugmentationWrapper:
    def __init__(self, augment_prob=0.3):
        self.augment_prob = augment_prob
        self.augment = transforms.Compose([
            transforms.RandomApply([
                transforms.ColorJitter(brightness=0.2, contrast=0.2, saturation=0.2, hue=0.05),
                transforms.RandomGrayscale(p=0.1),
                transforms.GaussianBlur(kernel_size=(3, 3), sigma=(0.3, 1.0)),
                transforms.RandomPerspective(distortion_scale=0.2, p=0.3),
            ], p=0.5),  # Less aggressive overall
            transforms.RandomHorizontalFlip(p=0.5),
            transforms.RandomRotation(degrees=10),
            transforms.RandomResizedCrop(size=80, scale=(0.9, 1.0), ratio=(0.95, 1.05)),
        ])

        self.normalize = transforms.Normalize(mean=[0.5, 0.5, 0.5],
                                              std=[0.5, 0.5, 0.5])

    def __call__(self, img):
        if random.random() < self.augment_prob:
            img = self.augment(img)
        return self.normalize(img)

class MFCCAugment:
    def __init__(self, 
                 p=0.4, 
                 time_mask_p=0.3, 
                 freq_mask_p=0.3, 
                 noise_p=0.3,
                 max_time_mask=10,
                 max_freq_mask=6,
                 noise_level=0.15):
        self.p = p
        self.time_mask_p = time_mask_p
        self.freq_mask_p = freq_mask_p
        self.noise_p = noise_p
        self.max_time_mask = max_time_mask
        self.max_freq_mask = max_freq_mask
        self.noise_level = noise_level

    def time_mask(self, mfcc):
        _, _, seq_len = mfcc.shape
        t = torch.randint(0, seq_len - self.max_time_mask, (1,))
        w = torch.randint(1, self.max_time_mask + 1, (1,))
        mfcc[:, :, t:t+w] = 0
        return mfcc

    def freq_mask(self, mfcc):
        _, num_mfcc, _ = mfcc.shape
        f = torch.randint(0, num_mfcc - self.max_freq_mask, (1,))
        w = torch.randint(1, self.max_freq_mask + 1, (1,))
        mfcc[:, f:f+w, :] = 0
        return mfcc

    def add_noise(self, mfcc):
        noise = self.noise_level * torch.randn_like(mfcc)
        return mfcc + noise

    def __call__(self, mfcc):
        if random.random() > self.p:
            return mfcc  # skip all augmentations
        
        if random.random() < self.time_mask_p:
            mfcc = self.time_mask(mfcc)
        if random.random() < self.freq_mask_p:
            mfcc = self.freq_mask(mfcc)
        if random.random() < self.noise_p:
            mfcc = self.add_noise(mfcc)
        
        return mfcc

class CustomDataset(Dataset):
    def __init__(self, dataset_path, mode="both", num_classes=None, device="cpu",augment=False):
        """
        mode: 'image', 'audio', or 'both'
        dataset_path: path to 'train' or 'dev' folder
        num_classes: total number of classes
        device: 'cpu' or 'cuda:0'
        """
        assert mode in {"image", "audio", "both"}
        self.samples = []
        self.mode = mode
        self.device = device
        self.augment = augment
        self.audio_augment = MFCCAugment()
        if(augment==True):
            self.image_transform = AugmentationWrapper(augment_prob=0.6)
        else:
            self.image_transform = transforms.Normalize(mean=[0.5, 0.5, 0.5],
                                                    std=[0.5, 0.5, 0.5])
        self.audio_transform = lambda waveform: waveform[:, 32000:]

        self.mfcc= T.MFCC(
                sample_rate=16_000,
                n_mfcc=13,
                log_mels=True,
                melkwargs={"n_fft": 400, "hop_length": 160, "n_mels": 23, "center": False})

        all_class_ids = sorted(os.listdir(dataset_path))
        if num_classes is None:
            self.num_classes = len(all_class_ids)
        else:
            self.num_classes = num_classes

        for class_id in all_class_ids:
            label = int(class_id) - 1  # start labels from 0
            class_path = os.path.join(dataset_path, class_id)

            img_files = sorted(glob(os.path.join(class_path, "*.png")))
            wav_files = sorted(glob(os.path.join(class_path, "*.wav")))

            img_dict = {os.path.splitext(os.path.basename(f))[0]: f for f in img_files}
            wav_dict = {os.path.splitext(os.path.basename(f))[0]: f for f in wav_files}

            shared_keys = sorted(img_dict.keys())
            if img_dict.keys() != wav_dict.keys():
                print(f"Warning: Image samples don't match audio samples in {class_id}")

            for key in shared_keys:
                img, audio = None, None

                if mode in {"image", "both"}:
                    img = read_image(img_dict[key]).float() / 255.0

                if mode in {"audio", "both"}:
                    waveform, freq = torchaudio.load(wav_dict[key])
                    assert freq == 16000
                    # throw out first two second of silence
                    audio = self.audio_transform(waveform)
                    mfccs = self.mfcc(audio)  # Shape [B, num_mfcc, T]
                    mean = mfccs.mean(dim=2, keepdim=True)
                    std = mfccs.std(dim=2, keepdim=True)

                    audio = (mfccs - mean) / (std + 1e-8)

                    # mean = audio.
                    # print(f"audio: {audio.shape}")
                    # print(f"mfcc: {mfcc_features.shape}")
                    

                if mode == "image":
                    self.samples.append((img, label))
                elif mode == "audio":
                    self.samples.append((audio, label))
                else:
                    self.samples.append(((img, audio), label))

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        item = self.samples[idx]
        
        if self.mode == "image":
            img, label = item
            img = self.image_transform(img)
            return img.to(self.device), self.one_hot(label).to(self.device)

        elif self.mode == "audio":
            audio, label = item
            if(self.augment == True):
                audio = self.audio_augment(audio)
            return audio.to(self.device), self.one_hot(label).to(self.device)

        else:
            (img, audio), label = item
            img = self.image_transform(img)
            if(self.augment == True):
                audio = self.audio_augment(audio)
            return (img.to(self.device), audio.to(self.device)), self.one_hot(label).to(self.device)

    def one_hot(self,label):
        zeros = torch.zeros(self.num_classes)
        zeros[label] = 1
        return zeros
