import torch
import torch.nn as nn
import torch.nn.functional as F

class ResidualBlock(nn.Module):
    def __init__(self, in_channels, out_channels, stride=1, dropout_p=0.3):
        super().__init__()
        self.same_shape = (in_channels == out_channels) and (stride == 1)
        # conv3x3 → BN → ReLU → Dropout → conv3x3 → BN → add shortcut → ReLU

        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1)
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.dropout = nn.Dropout2d(p=dropout_p)

        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1)
        self.bn2 = nn.BatchNorm2d(out_channels)

        if not self.same_shape:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride),
                nn.BatchNorm2d(out_channels)
            )
        else:
            self.shortcut = nn.Identity()

    def forward(self, x):
        identity = self.shortcut(x)
        out = self.relu(self.bn1(self.conv1(x)))
        out = self.dropout(out)
        out = self.bn2(self.conv2(out))
        return self.relu(out + identity)

# Inspired by Resnet18
# Using stride 2, to halve the resolution instead of pooling
class ImageModel(nn.Module):
    def __init__(self, dropout_p=0.3,batch_norm=True):
        super().__init__()
        self.output_dim = 256

        # Prepare before the resnets
        self.stem = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True)
        )

        # prepare two res blocks
        # first one with steps 2, halfing the size
        def make_stage(in_ch, out_ch, num_blocks, stride):
            layers = [ResidualBlock(in_ch, out_ch, stride=stride, dropout_p=dropout_p)]
            for _ in range(num_blocks - 1):
                layers.append(ResidualBlock(out_ch, out_ch, stride=1, dropout_p=dropout_p))
            return nn.Sequential(*layers)

        self.layer1 = make_stage(64, 128, num_blocks=2, stride=2)   # 80 → 40
        self.layer2 = make_stage(128, 256, num_blocks=2, stride=2)  # 40 → 20
        self.layer3 = make_stage(256, 512, num_blocks=2, stride=2)  # 20 → 10
        self.layer4 = make_stage(512, 512, num_blocks=2, stride=2)  # 10 → 5

        self.pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512, self.output_dim)

        self._init_weights()

    def _init_weights(self):
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.kaiming_normal_(m.weight, nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.stem(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.pool(x)
        x = x.view(x.size(0), -1)
        return self.fc(x)
