import torch
import torch.nn as nn
import torch.nn.functional as F
import torchaudio.transforms as T
from AudioModels import XVectorModel
from ImageModels import ImageModel
from aam_loss import ArcFace
from torch.nn.functional import normalize 

#soft attention fusion
class SoftAttention(nn.Module):
    def __init__(self,audio_dim,image_dim,out_dim):
        super(SoftAttention, self).__init__()
        #fusion part
        self.audio_transform = nn.Linear(audio_dim, out_dim)
        self.image_transform= nn.Linear(image_dim, out_dim)

        self.cross_modality_attention = nn.Linear(image_dim+audio_dim,2)

    def forward(self,audio_input,video_input):
        audio_input = normalize(audio_input)
        video_input = normalize(video_input)


        multi_mod_attention = self.cross_modality_attention(torch.cat((audio_input,video_input),dim=1))
        multi_mod_attention = nn.functional.softmax(multi_mod_attention, dim=1)
        
        audio_emb_transformed = self.audio_transform(audio_input)
        face_emb_transformed = self.image_transform(video_input)


        weighted_audio_emb = audio_emb_transformed * multi_mod_attention[:,0].unsqueeze(1)
        weighted_face_emb = face_emb_transformed * multi_mod_attention[:,1].unsqueeze(1)
        emb = torch.add(weighted_audio_emb,weighted_face_emb)
        return emb

class SoftMaxCrossEntropyLoss(nn.Module):
    def __init__(self,input_dimension,number_of_classes):
        super(SoftMaxCrossEntropyLoss, self).__init__()

        self.lin_layer = nn.Linear(input_dimension,number_of_classes,bias=False)
        torch.nn.init.xavier_normal_(self.lin_layer.weight)

        self.loss_func = nn.CrossEntropyLoss()
    
    def forward(self,inputs,labels):
        predictions = self.lin_layer(inputs)

        loss = self.loss_func(predictions,labels)
        return loss,predictions


class CombinedModel(nn.Module):
    def __init__(self, image_model, audio_model):
        super(CombinedModel, self).__init__()
        self.image_model = image_model
        self.audio_model = audio_model

    def forward(self, image, audio):
        img_emb = self.image_model(image)  # [B, 256]
        aud_emb = self.audio_model(audio)  # [B, 256]
        return img_emb, aud_emb

class TrainingModel(nn.Module):
    def __init__(self, num_classes, mode="both",dropout_p=0.1,batch_norm=True):
        super(TrainingModel, self).__init__()
        self.mode = mode
        if(mode == "audio"):
            self.model = XVectorModel(dropout_p=dropout_p,batch_norm=batch_norm)
        elif(mode == "image"):
            self.model = ImageModel(dropout_p=dropout_p,batch_norm=batch_norm)
        else:
            self.model = CombinedModel(ImageModel(dropout_p=dropout_p,batch_norm=batch_norm),XVectorModel(dropout_p=dropout_p,batch_norm=batch_norm))
    
        self.audio_loss_fc = ArcFace(256,num_classes,scale=30,margin=0.5)
        self.image_loss_fc = ArcFace(256,num_classes,scale=30,margin=0.5)
        if(mode == "both"):
            self.loss_fc = ArcFace(256,num_classes,scale=30,margin=0.5)
            self.fusion = SoftAttention(256,256,256)

    def forward(self,data,labels):
        labels = torch.argmax(labels, dim=1)#conver one-hot labels to index labels
        if(self.mode == "image"):
            embeddings = self.model(data) #get the output embeddings
            loss,predictions = self.image_loss_fc(embeddings,labels)#compute losses
        elif(self.mode == "audio"):
            embeddings = self.model(data) #get the output embeddings
            loss,predictions = self.audio_loss_fc(embeddings,labels)#compute losses
        else:
            image_data, audio_data = data
            img_emb, aud_emb = self.model(image_data, audio_data)
            embeddings = self.fusion(img_emb,aud_emb)
            loss,predictions = self.loss_fc(embeddings,labels)#compute losses

            return loss,predictions
        return loss,predictions
